% File: mhmm_test.m

close all

%% Parameters
image_file = 'images/sub/waldo-cal-01.jpg';
test_file = 'images/sub/red-cal-01.jpg';
sub_window = [5 40];
num_dct_coeffs = 3;
G = 2;                  % Number of Gaussians
M = 3;                  % Number of states
D = num_dct_coeffs * 3; % Number of dimensions (coeffs x colors)

% Define regions from picture (states)
obs_head = 1:1:18;
obs_torso = 19:1:35;
obs_hips = 36:1:46;

% Define initial HMM parameters

pi0 = [1 0 0]';
A0 =   [0.5 0.5  0  ; ...
         0  0.5 0.5 ; ...
         0   0   1  ];
        

%% Train HMM

% Read image
im_info = imfinfo(image_file);
im = imread(image_file);
im = double(im)/256;

% Display original image
figure
imshow(im);

% Calculate DCT for each window by scanning down the image
obs_length = im_info.Height - sub_window(1) + 1;
obs = zeros(obs_length, num_dct_coeffs * 3);
for t = 1:1:obs_length
    
    % Find DCT of red, green, blue separately
    im_dct_r = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),1));
    im_dct_g = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),2));
    im_dct_b = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),3));

    % Extract desired coefficients from DCT
    dct_c_r = extract_dct_coeff(im_dct_r, num_dct_coeffs);
    dct_c_g = extract_dct_coeff(im_dct_g, num_dct_coeffs);
    dct_c_b = extract_dct_coeff(im_dct_b, num_dct_coeffs);

    % Combine DCT coefficients into one observation vector
    dct_c = [dct_c_r dct_c_g dct_c_b];
    obs(t,:) = dct_c;
    
end

% Plot observation sequence
figure
surf(obs);

% Exctract test region for single dimension
% tobs = obs(obs_head,1);
% figure
% plot(tobs)
% figure
% [hist_tobs, xhist] = hist(tobs);
% bar(xhist, hist_tobs/sum(hist_tobs))
% hold on
% gmm_obj = gmdistribution.fit(tobs,3);
% plot(0:0.01:11,pdf(gmm_obj,[0:0.01:11]'))
% hold off
% disp(['Negative Log Likelihood: ', num2str(gmm_obj.NlogL)])

% Extract head region gmm for all dimensions
tobs = obs(obs_head,:);
gmm_head = gmdistribution.fit(tobs,G,'CovType','diagonal');
disp(['NlogL of Head: ', num2str(gmm_head.NlogL)])

% Extract torso region gmm for all dimensions
tobs = obs(obs_torso,:);
gmm_torso = gmdistribution.fit(tobs,G,'CovType','diagonal');
disp(['NlogL of Torso: ', num2str(gmm_torso.NlogL)])

% Extract hips region gmm for all dimensions
tobs = obs(obs_hips,:);
gmm_hips = gmdistribution.fit(tobs,G,'CovType','diagonal');
disp(['NlogL of Hips: ', num2str(gmm_hips.NlogL)])

% Construct mu matrix for means of GMMs
mu0 = zeros(D, M, G);
mu0(:,1,:) = gmm_head.mu';
mu0(:,2,:) = gmm_torso.mu';
mu0(:,3,:) = gmm_hips.mu';

% Construct sigma matrix for covariances of GMMs
sigma0 = zeros(D, D, M, G);
for i = 1:1:G
    sigma0(:,:,1,i) = diag(gmm_head.Sigma(:,:,i));
    sigma0(:,:,2,i) = diag(gmm_torso.Sigma(:,:,i));
    sigma0(:,:,3,i) = diag(gmm_hips.Sigma(:,:,i));
end

% Construct mixture matrix for each of the states
mixmat0 = zeros(M, G);
mixmat0(1,:) = gmm_head.PComponents;
mixmat0(2,:) = gmm_torso.PComponents;
mixmat0(3,:) = gmm_hips.PComponents;

% Train HMM using BWA (EM)
T = obs_length;
obs = obs';
[logliks, pi, A, mu, sigma, mixmat] = ...
    mhmm_em(obs, pi0, A0, mu0, sigma0, mixmat0, 'max_iter', 100);
figure
plot(logliks);

%% Test image

% Read test image
im_info = imfinfo(test_file);
im = imread(test_file);
im = double(im)/256;

% Display test image
figure
imshow(im);

% Calculate DCT for each window by scanning down the image
obs_length = im_info.Height - sub_window(1) + 1;
obs = zeros(obs_length, num_dct_coeffs * 3);
for t = 1:1:obs_length
    
    % Find DCT of red, green, blue separately
    im_dct_r = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),1));
    im_dct_g = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),2));
    im_dct_b = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),3));

    % Extract desired coefficients from DCT
    dct_c_r = extract_dct_coeff(im_dct_r, num_dct_coeffs);
    dct_c_g = extract_dct_coeff(im_dct_g, num_dct_coeffs);
    dct_c_b = extract_dct_coeff(im_dct_b, num_dct_coeffs);

    % Combine DCT coefficients into one observation vector
    dct_c = [dct_c_r dct_c_g dct_c_b];
    obs(t,:) = dct_c;
    
end

% Plot observation sequence from test image
figure
surf(obs);

% Calculate P[O|HMM] using Forward-Backward Algorithm
obs = obs';
[loglik, errors] = mhmm_logprob(obs, pi, A, mu, sigma, mixmat);
disp(['Log likelihood: ', num2str(loglik)])