% File: read_mhmm.m
% Read in Gaussian mixture HMM parameters from file
% Syntax:
%   [A,Pi,Mu,Sigma,W,N,D,G] = read_mhmm(file)

% This reads in the parameters (A, Pi, Mu, Sigma, W) from a HMM file
% dealing with continuous valued Gaussian mixture models
%
% Author: Shawn Hymel
% Date: December 9, 2011
% ECE 5554 - Computer Vision
%
% Input:
%   file        - file location of HMM parameters
% Output:
%   A           - estimated A matrix (state transitions)
%   Pi          - estimated Pi matrix (initial state probabilities)
%   Mu          - estimated Mu matrix (mean array)
%   Sigma       - estimated Sigma matrix (covariance matrices)
%   W           - estimated W matrix (mixture components)
%   N           - number of states
%   D           - number of dimensions for each observation vector
%   G           - number of Gaussians in each state

function [A,Pi,Mu,Sigma,W,N,D,G] = read_mhmm(file)

% Open the configuration file and read in the parameters
fid = fopen(file, 'r');
section = 0;
while 1
    
    % Read a line - if end of file, break loop
    tline = fgetl(fid);
    if ~ischar(tline) 
        break
    end
    
    % If empty line or '#' is found, skip line
    if (length(tline) <= 0 || tline(1) == '#')
        continue;
    end
        
    % First section - number of states
    if section == 0
        N = str2num(tline);
        A = zeros(N, N);
        
    % Second section - number of dimensions
    elseif section == 1
        D = str2num(tline);
        
    % Third section - number of Gaussians
    elseif section == 2
        G = str2num(tline);
        Mu = zeros(D, N, G);
        Mu_t = zeros(D * G, N);
        Sigma = zeros(D, D, N, G);
        Sigma_t = zeros(D * N * G, D);
        W = zeros(N, G);
        
    % Fourth section - initial state probabilities
    elseif section == 3
        Pi = str2num(tline);
        
    % Fifth section - A matrix
    elseif section <= 3 + N
        A(section - 3,:) = str2num(tline);
        
    % Sixth section - Mu matrix
    elseif section <= 3 + N + (D*G)
        Mu_t(section - 3 - N, :) = str2num(tline);
    
    % Seventh section - Sigma matrix
    elseif section <= 3 + N + (D*G) + (D*N*G)
        Sigma_t(section - 3 - N - (D*G), :) = str2num(tline);
        
    elseif section <= 3 + N + (D*G) + (D*N*G) + N
        W(section - 3 - N - (D*G) - (D*N*G), :) = str2num(tline);
    end   

    % Increment section counter
    section = section + 1;
    
end
    
% Close file
fclose(fid);

% Reshape Mu
for i = 1:1:G
    Mu(:,:,i) = Mu_t((((i-1)*D)+1):1:(i*D),:);
end
clear Mu_t

% Reshape Sigma
for i = 1:1:G
    for j = 1:1:N
        section = ((i-1)*N) + (j-1);
        Sigma(:,:,j,i) = Sigma_t(((section*D)+1):1:((section*D)+D),:);
    end
end
clear Sigma_t

% End function