% File: train_image_mhmm.m

% Author: Shawn Hymel
% Date: December 9, 2011
% ECE 5554 - Computer Vision

% This function reads a sample sub-image, calculates the necessary DCTs of
% each section and trains a continuous-valued HMM with Gaussian mixtures.
% The HMM parameters are then stored in a text file.

close all
clear all

%% Parameters

% Define user parameters
image_file = 'images/sub/waldo-cal-01-lq.jpg';
hmm_file = 'models/hmm_waldo.txt';
sub_window = [5 20];    % Size of sampling window to construct obs vector
num_dct_coeffs = 3;     % Number of DCT coefficients to extract
G = 2;                  % Number of Gaussians
N = 3;                  % Number of states
D = num_dct_coeffs * 3; % Number of dimensions (coeffs x colors)

% Define regions from picture (states)
obs_hat = 1:1:10;
obs_face = 11:1:20;
obs_shirt = 21:1:36;

% Define initial HMM parameters (left-to-right, linear model)
pi0 = [1 0 0]';
A0 =  [ 0.5 0.5  0  ; ...
         0  0.5 0.5 ; ...
         0   0   1 ];
     
%% Train HMM

% Read image
im_info = imfinfo(image_file);
im = imread(image_file);
im = double(im)/256;

% Display original image
figure
imshow(im);

% Calculate DCT for each window by scanning down the image
obs_length = im_info.Height - sub_window(1) + 1;
obs = zeros(obs_length, num_dct_coeffs * 3);
for t = 1:1:obs_length
    
    % Find DCT of red, green, blue separately
    im_dct_r = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),1));
    im_dct_g = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),2));
    im_dct_b = dct2(im(t:t+sub_window(1)-1,1:sub_window(2),3));

    % Extract desired coefficients from DCT
    dct_c_r = extract_dct_coeff(im_dct_r, num_dct_coeffs);
    dct_c_g = extract_dct_coeff(im_dct_g, num_dct_coeffs);
    dct_c_b = extract_dct_coeff(im_dct_b, num_dct_coeffs);

    % Combine DCT coefficients into one observation vector
    dct_c = [dct_c_r dct_c_g dct_c_b];
    obs(t,:) = dct_c;
    
end

% Plot observation sequence
figure
surf(obs);

% Explain log likelihood
disp(' ')
disp('The Negative Log Likelihood of each region is displayed below.')
disp('Ensure that each of the numbers is POSITIVE for a good GMM fit.')
disp('Adjust parameters in script to generate small, positive numbers.')

% Extract hat region GMM for all dimensions
tobs = obs(obs_hat,:);
gmm_hat = gmdistribution.fit(tobs, G, 'CovType', 'diagonal');
disp(['NlogL of Hat: ', num2str(gmm_hat.NlogL)])

% Extract face region GMM for all dimensions
tobs = obs(obs_face,:);
gmm_face = gmdistribution.fit(tobs, G, 'CovType', 'diagonal');
disp(['NlogL of Face: ', num2str(gmm_face.NlogL)])

% Extract shirt region GMM for all dimensions
tobs = obs(obs_shirt,:);
gmm_shirt = gmdistribution.fit(tobs, G, 'CovType', 'diagonal');
disp(['NlogL of Shirt: ', num2str(gmm_shirt.NlogL)])
disp(' ')

% Construct mu matrix for means of GMMs
mu0 = zeros(D, N, G);
mu0(:,1,:) = gmm_hat.mu';
mu0(:,2,:) = gmm_face.mu';
mu0(:,3,:) = gmm_shirt.mu';

% Construct sigma matrix for covariances of GMMs
sigma0 = zeros(D, D, N, G);
for i = 1:1:G
    sigma0(:,:,1,i) = diag(gmm_hat.Sigma(:,:,i));
    sigma0(:,:,2,i) = diag(gmm_face.Sigma(:,:,i));
    sigma0(:,:,3,i) = diag(gmm_shirt.Sigma(:,:,i));
end

% Construct mixture matrix for each of the states
mixmat0 = zeros(N, G);
mixmat0(1,:) = gmm_hat.PComponents;
mixmat0(2,:) = gmm_face.PComponents;
mixmat0(3,:) = gmm_shirt.PComponents;

% Train HMM using Baum-Welch Algorithm (Expectation Maximization)
T = obs_length;
obs = obs';
disp('Performing BWA. Ensure that the Log Likelihood converges.')
[logliks, pi, A, mu, sigma, mixmat] = ...
    mhmm_em(obs, pi0, A0, mu0, sigma0, mixmat0, 'max_iter', 100);
figure
plot(logliks);
disp(' ')

%% Record HMM to a File

% Write contents to a file
obs_y = length(obs);
obs_x = 1;
fid = fopen(hmm_file, 'w');
fprintf(fid, '# MATLAB generated parameters for continuous-value HMM\n\n');
fprintf(fid, '# Number of states (N)\n');
fprintf(fid, '%i\n\n', N);
fprintf(fid, '# Number of dimensions (D)\n');
fprintf(fid, '%i\n\n', D);
fprintf(fid, '# Number of Gaussians (G)\n');
fprintf(fid, '%i\n\n', G);
fprintf(fid, '# Initial state probability matrix (Pi)\n');
fprintf(fid, '%f ', pi);
fprintf(fid, '\n\n');
fprintf(fid, '# State transition probability matrix (A)\n');
for i = 1:1:N
    fprintf(fid, '%f ', A(i,:));
    fprintf(fid, '\n');
end
fprintf(fid, '\n');
fprintf(fid, '# Mean array (Mu)\n');
for i = 1:1:G
    for j = 1:1:D
        fprintf(fid, '%f ', mu(j,:,i));
        fprintf(fid, '\n');
    end
end
fprintf(fid, '\n');
fprintf(fid, '# Covariance matrices (Sigma)\n');
for i = 1:1:G
    for j = 1:1:N
        for k = 1:1:D
            fprintf(fid, '%f ', sigma(k,:,j,i));
            fprintf(fid, '\n');
        end
    end
end
fprintf(fid, '\n');
fprintf(fid, '# Mixture matrix (W)\n');
for i = 1:1:N
    fprintf(fid, '%f ', mixmat(i,:));
    fprintf(fid, '\n');
end
fclose(fid);